"""
Extension for the `TaskCog`.
Loading this file via `commands.Bot.load_extension` will add `TaskCog` to the bot.
"""

import re
import json
import logging
import datetime
from os import getenv
from time import process_time

import aiohttp
from discord import TextChannel, Embed, Colour
from discord import app_commands
from discord.ext import commands, tasks
from discord.errors import Forbidden
from feedparser import parse

from feed import RSSFeed, Subscription, RSSItem
from utils import get_unparsed_feed
from filters import match_text
from api import API

log = logging.getLogger(__name__)

TASK_INTERVAL_MINUTES = getenv("TASK_INTERVAL_MINUTES")
subscription_task_times = [
    datetime.time(hour, minute, tzinfo=datetime.timezone.utc)
    for hour in range(24)
    for minute in range(0, 60, int(TASK_INTERVAL_MINUTES))
]
log.debug("Task will trigger every %s minutes", TASK_INTERVAL_MINUTES)


class TaskCog(commands.Cog):
    """
    Tasks cog for PYRSS.
    """

    def __init__(self, bot):
        super().__init__()
        self.bot = bot

    @commands.Cog.listener()
    async def on_ready(self):
        """
        Instructions to execute when the cog is ready.
        """
        self.subscription_task.start()
        log.info("%s cog is ready", self.__class__.__name__)

    @commands.Cog.listener(name="cog_unload")
    async def on_unload(self):
        """
        Instructions to execute before the cog is unloaded.
        """
        self.subscription_task.cancel()

    @app_commands.command(name="debug-trigger-task")
    async def debug_trigger_task(self, inter):
        await inter.response.defer()
        await self.subscription_task()
        await inter.followup.send("done")

    @tasks.loop(time=subscription_task_times)
    async def subscription_task(self):
        """
        Task for fetching and processing subscriptions.
        """
        log.info("Running subscription task")

        async with aiohttp.ClientSession() as session:
            api = API(self.bot.api_token, session)
            subscriptions = await self.get_subscriptions(api)
            await self.process_subscriptions(api, subscriptions)

    async def get_subscriptions(self, api: API) -> list[Subscription]:
        guild_ids = [guild.id for guild in self.bot.guilds]
        sub_data = []

        for page, _ in enumerate(iter(int, 1)):
            try:
                log.debug("fetching page '%s'", page + 1)
                sub_data.extend(
                    (await api.get_subscriptions(server__in=guild_ids, page=page+1))[0]
                )
            except aiohttp.ClientResponseError as error:
                match error.status:
                    case 404:
                        log.debug("final page reached '%s'", page)
                        break
                    case 403:
                        log.critical(error)
                        self.subscription_task.cancel()
                        return []  # returning an empty list should gracefully end the task
                    case _:
                        log.error(error)
                        break

            except Exception as error:
                log.error("Exception while gathering page data %s", error)
                break


        return Subscription.from_list(sub_data)

    async def process_subscriptions(self, api: API, subscriptions: list[Subscription]):
        for sub in subscriptions:
            log.debug("processing subscription '%s'", sub.id)

            if not sub.active or not sub.channels_count:
                continue

            unparsed_feed = await get_unparsed_feed(sub.url, api.session)
            parsed_feed = parse(unparsed_feed)

            rss_feed = RSSFeed.from_parsed_feed(parsed_feed)
            await self.process_items(api, sub, rss_feed)

    async def process_items(self, api: API, sub: Subscription, feed: RSSFeed):
        log.debug("processing items")

        channels = [self.bot.get_channel(channel.channel_id) for channel in await sub.get_channels(api)]
        filters = [await api.get_filter(filter_id) for filter_id in sub.filters]

        log.debug(json.dumps(filters, indent=4))

        for item in feed.items:
            log.debug("processing item '%s'", item.guid)

            if item.pub_date < sub.published_threshold:
                log.debug("item '%s' older than subscription threshold '%s', skipping", item.pub_date, sub.published_threshold)
                continue

            log.debug("before")
            blocked = any(self.filter_item(_filter, item) for _filter in filters)
            log.debug("after")
            mutated_item = item.create_mutated_copy(sub.mutators)

            for channel in channels:
                successful_track = await self.mark_tracked_item(api, sub, item, channel.id, blocked)

                if successful_track and not blocked:
                    await channel.send(embed=await mutated_item.to_embed(sub, feed, api.session))

    def filter_item(self, _filter: dict, item: RSSItem) -> bool:
        """
        Returns `True` if item should be ignored due to filters.
        """

        log.debug("checking filter")

        try:
            match_found = match_text(_fliter, item.title) or match_text(_filter, item.description)
            log.debug("filter match found? '%s'", match_found)
            return match_found
        except Exception as error:
            log.error(error)
            input("[paused] >")
            return False

    async def mark_tracked_item(self, api: API, sub: Subscription, item: RSSItem, channel_id: int, blocked: bool):
        try:
            log.debug("marking as tracked 'blocked: %s'", blocked)
            await api.create_tracked_content(
                guid=item.guid,
                title=item.title,
                url=item.link,
                subscription=sub.id,
                channel_id=channel_id,
                blocked=blocked
            )
            return True
        except aiohttp.ClientResponseError as error:
            if error.status == 409:
                log.debug(error)
            else:
                log.error(error)

            return False


async def setup(bot):
    """
    Setup function for this extension.
    Adds `TaskCog` to the bot.
    """

    cog = TaskCog(bot)
    await bot.add_cog(cog)
    log.info("Added %s cog", cog.__class__.__name__)
